TOPIC_HIERARCHY_FIELDS = [
    "organization",
    "system_id",
    "event_type",
    "subsystem_id",
    "device_id",
]

# Payload fields
ID = "id"
TIME = "time"
SOURCE = "source"
TYPE = "type"
DATA = "data"
METADATA = "metadata"
SERIAL_NUMBER = "serial_number"
